// Plural forms package
// (c) Peter Kankowski, 2008-2011. http://www.strchr.com/plural_forms kankowski@narod.ru

#define _CRT_SECURE_NO_DEPRECATE // for sscanf function
#include <string.h>
#include <stdio.h>
#include "plurals.h"

int inline InRange(unsigned n, unsigned start, unsigned end) { // start <= n <= end
	// Compatibility: two's complement machines only
	return (unsigned)(n - start) <= (unsigned)(end - start);
}

unsigned PluralsReadCfg(PLURAL_INFO* info, const char* desc) {
	memset(info->form, 0, sizeof(info->form));
	info->nforms = 0;
	info->zero_form = (unsigned)~0;

	if(1 == sscanf(desc, " z = %d ", &info->zero_form))
		desc += strcspn(desc, ";") + 1;

	for(unsigned i = 0; i < MAX_FORMS; i++) {
		int fields = sscanf(desc, " %d %d %d %1[^;] ",
			&info->form[info->nforms].start,
			&info->form[info->nforms].finish,
			&info->form[info->nforms].modulo,
			&info->form[info->nforms].skip_teens);
		if(EOF == fields)
			break;
		if(1 == fields) // "1" is equivalent to "1 1"
			info->form[info->nforms].finish = info->form[info->nforms].start;
		if(0 != fields) // skip invalid forms
			info->nforms++;

		const char* next = desc + strcspn(desc, ";");
		if('\0' == *next)
			break;
		desc = next + 1;
	}
	return info->nforms + 1;
}

unsigned PluralsGetForm(const PLURAL_INFO* info, unsigned n) {
	// Use zero form if it's valid
	if(n == 0 && info->zero_form <= info->nforms)
		return info->zero_form;

	for(unsigned i = 0; i < info->nforms; i++) {
		unsigned x = n;
		if(info->form[i].modulo != 0)
			x = n % info->form[i].modulo;

		if(info->form[i].skip_teens && InRange(n % 100, 10, 19))
			continue;
		if(InRange(x, info->form[i].start, info->form[i].finish))
			return i;
	}
	// The default plural form
	return info->nforms;
}
